//********************************************************* 
// 
// Copyright (c) Microsoft. All rights reserved. 
// This code is licensed under the MIT License (MIT). 
// THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF 
// ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING ANY 
// IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR 
// PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT. 
// 
//*********************************************************

#pragma once

// Level:
// This is an abstract class from which all of the levels of the game are derived.
// Each level potentially overrides up to four methods:
//     Initialize - (required) takes a list of objects and enables the objects that
//         are active for the level as well as setting their positions and
//         any animations associated with the objects.
//     Update - this method is called once per time step and is expected to
//         determine if the level has been completed. The Level class provides
//         a 'standard' Update method which checks each object that is a target
//         and disables any active targets that have been hit. It returns true
//         once there are no active targets remaining.
//     SaveState - method to save any Level specific state. Default is defined as
//         not saving any state.
//     LoadState - method to restore any Level specific state. Default is defined
//         as not restoring any state.

#include "GameObject.h"
#include "PersistentState.h"

class Level
{
public:
    virtual void Initialize(
        std::vector<std::shared_ptr<GameObject>> const& objects
        ) = 0;

    virtual bool Update(
        float time,
        float elapsedTime,
        float timeRemaining,
        std::vector<std::shared_ptr<GameObject>> const& objects
        );

    virtual void SaveState(PersistentState& state);
    virtual void LoadState(PersistentState& state);

    std::wstring const& Objective();
    float TimeLimit();

protected:
    std::wstring   m_objective;
    float          m_timeLimit;
};