//********************************************************* 
// 
// Copyright (c) Microsoft. All rights reserved. 
// This code is licensed under the MIT License (MIT). 
// THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF 
// ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING ANY 
// IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR 
// PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT. 
// 
//*********************************************************

#include "pch.h"
#include "SphereMesh.h"
#include "ConstantBuffers.h"

using namespace DirectX;

SphereMesh::SphereMesh(_In_ winrt::com_ptr<ID3D11Device3> const& device, uint16_t segments)
{
    D3D11_BUFFER_DESC bd = { 0 };
    D3D11_SUBRESOURCE_DATA initData = { 0 };

    uint16_t slices = segments / 2;
    uint16_t numVertices = (slices + 1) * (segments + 1) + 1;
    uint16_t numIndices = slices * segments * 3 * 2;

    std::vector<PNTVertex> point(numVertices);
    std::vector<uint16_t> index(numIndices);

    // To make the texture look right on the top and bottom of the sphere
    // each slice will have 'segments + 1' vertices. The top and bottom
    // vertices will all be coincident, but have different U texture cooordinates.
    uint16_t p = 0;
    for (uint16_t a = 0; a <= slices; a++)
    {
        float angle1 = static_cast<float>(a) / static_cast<float>(slices) * XM_PI;
        float z = cos(angle1);
        float r = sin(angle1);
        for (uint16_t b = 0; b <= segments; b++)
        {
            float angle2 = static_cast<float>(b) / static_cast<float>(segments) * XM_2PI;
            point[p].position = XMFLOAT3(r * cos(angle2), r * sin(angle2), z);
            point[p].normal = point[p].position;
            point[p].textureCoordinate = XMFLOAT2((1.0f - z) / 2.0f, static_cast<float>(b) / static_cast<float>(segments));
            p++;
        }
    }
    m_vertexCount = p;

    p = 0;
    for (uint16_t a = 0; a < slices; a++)
    {
        uint16_t p1 = a * (segments + 1);
        uint16_t p2 = (a + 1) * (segments + 1);

        // Generate two triangles for each segment around the slice.
        for (uint16_t b = 0; b < segments; b++)
        {
            if (a < (slices - 1))
            {
                // For all but the bottom slice add the triangle with one
                // vertex in the a slice and two vertices in the a + 1 slice.
                // Skip it for the bottom slice since the triangle would be
                // degenerate as all the vertices in the bottom slice are coincident.
                index[p] = b + p1;
                index[p + 1] = b + p2;
                index[p + 2] = b + p2 + 1;
                p = p + 3;
            }
            if (a > 0)
            {
                // For all but the top slice add the triangle with two
                // vertices in the a slice and one vertex in the a + 1 slice.
                // Skip it for the top slice since the triangle would be
                // degenerate as all the vertices in the top slice are coincident.
                index[p] = b + p1;
                index[p + 1] = b + p2 + 1;
                index[p + 2] = b + p1 + 1;
                p = p + 3;
            }
        }
    }
    m_indexCount = p;

    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = sizeof(PNTVertex) * m_vertexCount;
    bd.BindFlags = D3D11_BIND_VERTEX_BUFFER;
    bd.CPUAccessFlags = 0;
    initData.pSysMem = point.data();
    winrt::check_hresult(
        device->CreateBuffer(&bd, &initData, m_vertexBuffer.put())
        );

    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = sizeof(uint16_t) * m_indexCount;
    bd.BindFlags = D3D11_BIND_INDEX_BUFFER;
    bd.CPUAccessFlags = 0;
    initData.pSysMem = index.data();
    winrt::check_hresult(
        device->CreateBuffer(&bd, &initData, m_indexBuffer.put())
        );
}
